/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_proxy_sequence.h
 *
 * Purpose:     proxy_sequence template class.
 *
 * Created:     10th September 2003
 * Updated:     10th September 2003
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _STLSOFT_INCL_H_STLSOFT_PROXY_SEQUENCE
#define _STLSOFT_INCL_H_STLSOFT_PROXY_SEQUENCE

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
#define _STLSOFT_VER_H_STLSOFT_PROXY_SEQUENCE_MAJOR       1
#define _STLSOFT_VER_H_STLSOFT_PROXY_SEQUENCE_MINOR       0
#define _STLSOFT_VER_H_STLSOFT_PROXY_SEQUENCE_REVISION    3
#define _STLSOFT_VER_H_STLSOFT_PROXY_SEQUENCE_EDIT        3
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !_STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

template<   ss_typename_param_k E
        ,   ss_typename_param_k V
        ,   ss_typename_param_k T
        >
class proxy_sequence
{
public:
    typedef E                           element_type;
    typedef V                           value_type;
    typedef T                           traits_type;
    typedef proxy_sequence<E, V, T>     class_type;

public:
	/// Constructs a proxy_sequence from a given range.
    proxy_sequence(element_type *first, element_type *last)
        : m_begin(first)
        , m_end(last)
    {}

/// \name Iteration
/// @{
public:
    class iterator
    {
        typedef proxy_sequence<E, V, T>                         sequence_type;
        typedef iterator                                        class_type;
    public:
#if defined(__STLSOFT_COMPILER_IS_MSVC)
        typedef E                                               element_type;
#else /* ? compiler */
        typedef ss_typename_type_k sequence_type::element_type  element_type;
#endif /* compiler */

    private:
#if defined(__STLSOFT_COMPILER_IS_GCC) || \
    defined(__STLSOFT_COMPILER_IS_MWERKS)
        friend class sequence_type;
#else /* ? compiler */
        friend /* class */ sequence_type;
#endif /* compiler */
        iterator(element_type *first, element_type *last)
            : m_begin(first)
            , m_end(last)
        {}
    public:
        iterator()
            : m_begin(NULL)
            , m_end(NULL)
        {}
        iterator(iterator const &rhs)
            : m_begin(rhs.m_begin)
            , m_end(rhs.m_end)
        {}

        iterator &operator =(iterator const &rhs)
        {
            m_begin =   rhs.m_begin;
            m_end   =   rhs.m_end;
        
            return *this;
        }

    public:
	    /// Pre-increment operator
        class_type &operator ++()
        {
            stlsoft_message_assert("Incrementing invalid iterator", m_begin != m_end);

            ++m_begin;

            return *this;
        }
	    /// Post-increment operator
        class_type operator ++(int)
        {
			class_type	r(*this);

			operator ++();

            return r;
        }
		/// Dereference to return a value at the current position of type V
        value_type operator *() const
        {
            return traits_type::make_value(*m_begin);
        }

	    /// Evaluates whether \c this and \c rhs are equivalent
        bool operator ==(class_type const &rhs) const
        {
            if(m_end == rhs.m_end)
            {
                return m_begin == rhs.m_begin;
            }
            else
            {
                return (m_begin == m_end) == (rhs.m_begin == rhs.m_end);
            }
        }
	    /// Evaluates whether \c this and \c rhs are not equivalent
        bool operator !=(class_type const &rhs) const
        {
            return !operator ==(rhs);
        }

    /// Members
    private:
        element_type    *m_begin;
        element_type    *m_end;
    };

    /// Begins the iteration
    ///
    /// \return An iterator representing the start of the sequence
    iterator    begin()
    {
        return iterator(m_begin, m_end);
    }
    /// Ends the iteration
    ///
    /// \return An iterator representing the end of the sequence
    iterator    end()
    {
        return iterator();
    }
/// @}

/// \name Attributes
/// @{
public:
    /// Indicates whether the sequence is empty
    bool empty() const
    {
        return m_begin == m_end;
    }
/// @}

// Members
private:
    element_type    *m_begin;
    element_type    *m_end;
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !_STLSOFT_INCL_H_STLSOFT_PROXY_SEQUENCE */

/* ////////////////////////////////////////////////////////////////////////// */
